@@ -108,6 +108,7 @@ corresponding monitors whenever the monitors are active.
 
 (defun exwm-randr-enable ()
   "Enable EXWM RandR support."
+  (message "`exwm-randr-enable' is obsolete. Use `exwm-randr-mode' instead.")
   (exwm-randr-mode 1))
 (make-obsolete 'exwm-randr-enable "Use `exwm-randr-mode' instead." "0.30")
 
@@ -305,11 +306,10 @@ Refresh when any CRTC/output changes."
         notify)
     (xcb:unmarshal evt data)
     (with-slots (subCode u) evt
-      (cl-case subCode
-        (xcb:randr:Notify:CrtcChange
-         (setq notify (slot-value u 'cc)))
-        (xcb:randr:Notify:OutputChange
-         (setq notify (slot-value u 'oc))))
+      (cond ((= subCode xcb:randr:Notify:CrtcChange)
+	     (setq notify (slot-value u 'cc)))
+	    ((= subCode xcb:randr:Notify:OutputChange)
+	     (setq notify (slot-value u 'oc))))
       (when notify
         (with-slots (timestamp) notify
           (when (> timestamp exwm-randr--last-timestamp)
